@@ -8,26 +8,29 @@ package Mac::PropertyList::SAX;
 
 =head1 SYNOPSIS
 
-See L<Mac::PropertyList>
+See L<Mac::PropertyList|Mac::PropertyList>
 
 =head1 DESCRIPTION
 
-L<Mac::PropertyList> is useful, but very slow on large files because it does
-XML parsing itself, intead of handing it off to a dedicated parser. This module
-uses L<XML::SAX::ParserFactory> to select a parser capable of doing the heavy
-lifting, reducing parsing time on large files by a factor of 30 or more.
+L<Mac::PropertyList|Mac::PropertyList> is useful, but very slow on large files
+because it does XML parsing itself, intead of handing it off to a dedicated
+parser. This module uses L<XML::SAX::ParserFactory|XML::SAX::ParserFactory> to
+select a parser capable of doing the heavy lifting, reducing parsing time on
+large files by a factor of 30 or more.
 
-This module does not replace L<Mac::PropertyList>: it depends on it for some
-package definitions and plist printing routines. You should, however, be able
-to replace all C<use L<Mac::PropertyList>> lines with C<use
-Mac::PropertyList::SAX>, without changing anything else, and notice an
-immediate improvement in performance on large input files.
+This module does not replace L<Mac::PropertyList|Mac::PropertyList>: it depends
+on it for some package definitions and plist printing routines. You should,
+however, be able to replace all C<use Mac::PropertyList>
+lines with C<use Mac::PropertyList::SAX>, without changing anything else, and
+notice an immediate improvement in performance on large input files.
 
-Performance will depend largely on the parser that L<XML::SAX::ParserFactory>
-selects for you. By default, L<XML::SAX::Expat> is used; to change the parser
-used, set the environment variable C<MAC_PROPERTYLIST_SAX_PARSER> to a value
-accepted by $XML::SAX::ParserPackage from L<XML::SAX::ParserFactory> (or set
-$XML::SAX::ParserPackage directly).
+Performance will depend largely on the parser that
+L<XML::SAX::ParserFactory|XML::SAX::ParserFactory> selects for you. By default,
+L<XML::SAX::Expat|XML::SAX::Expat> is used; to change the parser used, set the
+environment variable C<MAC_PROPERTYLIST_SAX_PARSER> to a value accepted by
+$XML::SAX::ParserPackage from
+L<XML::SAX::ParserFactory|XML::SAX::ParserFactory> (or set
+C<$XML::SAX::ParserPackage> directly).
 
 =cut
 
@@ -43,8 +46,6 @@ use XML::SAX::ParserFactory;
 
 use base qw(Exporter);
 
-our $ENCODE_ENTITIES = 1;
-
 our @EXPORT_OK = qw(
     parse_plist 
     parse_plist_fh
@@ -62,10 +63,50 @@ our %EXPORT_TAGS = (
     parse  => [ qw(parse_plist parse_plist_fh parse_plist_file parse_plist_string) ],
 );
 
-our $VERSION = '0.84';
+our $VERSION = '0.85';
+
+
+
+=head1 CLASS VARIABLES
+
+Class scoped variables that control the packages settings.
+
+=over 4
+
+=item ENCODE_ENTITIES
+
+Allows the XHTML encoding of the data to be turned off. Default = C<1>
+
+=item ENCODE_UNSAFE_CHARS
+
+A Perl character class definition containing the only characters to be
+XHTML encoded. See HTML::Entities::encode_entities for description of
+the $unsafe_chars parameter. Default = C<undef>
+
+=cut
+
+our $ENCODE_ENTITIES     = 1;
+our $ENCODE_UNSAFE_CHARS = undef;
+
+=item OLD_BEHAVIOR
+
+Restores the old behavior of double encoding output data. Default = C<0>
+
+=cut
+
+our $OLD_BEHAVIOR = 0;
+
+=item XML::SAX::ParserPackage
+
+Parser to use. Can also be set with environment variable
+C<MAC_PROPERTYLIST_SAX_PARSER>. Default = C<"XML::SAX::Expat">
+
+=cut
 
 $XML::SAX::ParserPackage = $ENV{MAC_PROPERTYLIST_SAX_PARSER} || "XML::SAX::Expat";
 
+=back
+
 =head1 EXPORTS
 
 By default, no functions are exported. Specify individual functions to export
@@ -132,8 +173,8 @@ sub _parse {
 Create a plist from an array or hash reference.
 
 The values of the hash can be simple scalars or references. Hash and array
-references are handled recursively, and L<Mac::PropertyList> objects are output
-correctly.  All other scalars are treated as strings (use L<Mac::PropertyList>
+references are handled recursively, and L<Mac::PropertyList|Mac::PropertyList> objects are output
+correctly. All other scalars are treated as strings (use L<Mac::PropertyList|Mac::PropertyList>
 objects to represent other types of scalars).
 
 Returns a string representing the reference in serialized plist format.
@@ -148,7 +189,7 @@ sub create_from_ref {
             my ($hash) = @_;
             Mac::PropertyList::SAX::dict->write_open,
                 (map { "\t$_" } map {
-                    Mac::PropertyList::SAX::dict->write_key(_escape($_)),
+                    Mac::PropertyList::SAX::dict->write_key($OLD_BEHAVIOR ? _escape($_) : $_),
                     _handle_value($hash->{$_}) } keys %$hash),
                 Mac::PropertyList::SAX::dict->write_close
         }
@@ -166,7 +207,7 @@ sub create_from_ref {
            if (UNIVERSAL::can($val, 'write')) { $val->write }
         elsif (UNIVERSAL::isa($val,  'HASH')) { _handle_hash ($val) }
         elsif (UNIVERSAL::isa($val, 'ARRAY')) { _handle_array($val) }
-        else { Mac::PropertyList::SAX::string->new(_escape($val))->write }
+        else { Mac::PropertyList::SAX::string->new($OLD_BEHAVIOR ? _escape($val) : $val)->write }
     }
 
     $Mac::PropertyList::XML_head .
@@ -176,7 +217,7 @@ sub create_from_ref {
 
 =item create_from_hash( HASH_REF )
 
-Provided for backward compatibility with L<Mac::PropertyList>: aliases
+Provided for backward compatibility with L<Mac::PropertyList|Mac::PropertyList>: aliases
 create_from_ref.
 
 =cut
@@ -185,7 +226,7 @@ create_from_ref.
 
 =item create_from_array( ARRAY_REF )
 
-Provided for backward compatibility with L<Mac::PropertyList>: aliases
+Provided for backward compatibility with L<Mac::PropertyList|Mac::PropertyList>: aliases
 create_from_ref.
 
 =cut
@@ -198,7 +239,13 @@ B<Internal use only.> Escapes illegal characters into XML entities.
 
 =cut
 
-sub _escape { name2hex_xml(hex2name(encode_entities_numeric(@_))) }
+sub _escape {
+    my $string = join("\n",grep(defined,@_));
+    $ENCODE_ENTITIES && 
+        return name2hex_xml(hex2name(encode_entities_numeric($string,
+                                                             $ENCODE_UNSAFE_CHARS)));
+    return $string;
+}
 
 package Mac::PropertyList::SAX::Handler;
 
@@ -324,13 +371,13 @@ package Mac::PropertyList::SAX::array;
 use base qw(Mac::PropertyList::array);
 package Mac::PropertyList::SAX::dict;
 use base qw(Mac::PropertyList::dict);
-sub write_key { "<key>" . Mac::PropertyList::SAX::_escape($_[1]) . "</key>" }
+sub write_key { "<key>" . (Mac::PropertyList::SAX::_escape($_[1]) || '') . "</key>" }
 package Mac::PropertyList::SAX::Scalar;
 use base qw(Mac::PropertyList::Scalar);
 sub write {
     $_[0]->write_open .
-    Mac::PropertyList::SAX::_escape($_[0]->value) .
-    $_[0]->write_close
+        (Mac::PropertyList::SAX::_escape($_[0]->value) || '') .
+            $_[0]->write_close
 }
 use overload '""' => sub { $_[0]->as_basic_data };
 package Mac::PropertyList::SAX::date;
@@ -370,38 +417,48 @@ __END__
 =head1 BUGS / CAVEATS
 
 Any sane XML parser you can find to use with this module will decode
-XHTML-encoded entities in the original property list; L<Mac::PropertyList>
-doesn't touch them. Also, your XML parser may convert accented/special
-characters into '\x{ff}' sequences; these are preserved in their original
-encoding by L<Mac::PropertyList>.
+XHTML-encoded entities in the original property list;
+L<Mac::PropertyList|Mac::PropertyList> doesn't touch them. Also, your XML
+parser may convert accented/special characters into '\x{ff}' sequences; these
+are preserved in their original encoding by
+L<Mac::PropertyList|Mac::PropertyList>.
 
 Before version 0.80 of this module, characters invalid in XML were not
 serialized properly from create_from_ref(); before version 0.82, they were not
 serialized properly in plist_as_string(). Thanks to Jon Connell for pointing
 out these problems.
 
-Unlike L<Mac::PropertyList> and old versions (< 0.60) of
+Unlike L<Mac::PropertyList|Mac::PropertyList> and old versions (< 0.60) of
 Mac::PropertyList::SAX, this module does not trim leading and trailing
-whitespace from plist elements.  The difference in behavior is thought to be
+whitespace from plist elements. The difference in behavior is thought to be
 rarely noticeable; in any case, I believe this module's current behavior is the
 more correct. Any documentation that covers this problem would be appreciated.
 
 The behavior of create_from_hash and create_from_array has changed: these
 functions (which are really just aliases to the new create_from_ref function)
 are now capable of recursively serializing complex data structures. That is:
-for inputs that L<Mac::PropertyList>'s create_from_* functions handled, the
-output should be the same, I<but> this module supports inputs that
-L<Mac::PropertyList> does not.
+for inputs that L<Mac::PropertyList|Mac::PropertyList>'s create_from_*
+functions handled, the output should be the same, I<but> this module supports
+inputs that L<Mac::PropertyList|Mac::PropertyList> does not.
 
 Before version 0.83, this module left the selection of a SAX-based parser
-entirely to the discretion of L<XML::SAX::ParserFactory>. Unfortunately, it
-seems impossible to guarantee that the parser returned even supports XML
-(L<XML::SAX::RTF> could be returned), so it has become necessary to select a
-parser by default: L<XML::SAX::Expat>, which is now part of the dependencies of
-this module. If you know you will use another parser of a specific name, you
-can force installation without L<XML::SAX::Expat> and always specify the parser
-you wish to use by setting $XML::SAX::ParserPackage or the
-MAC_PROPERTYLIST_SAX_PARSER environment variable (see L<DESCRIPTION>).
+entirely to the discretion of
+L<XML::SAX::ParserFactory|XML::SAX::ParserFactory>. Unfortunately, it seems
+impossible to guarantee that the parser returned even supports XML
+(L<XML::SAX::RTF|XML::SAX::RTF> could be returned), so it has become necessary
+to select a parser by default: L<XML::SAX::Expat|XML::SAX::Expat>, which is now
+part of the dependencies of this module. If you know you will use another
+parser of a specific name, you can force installation without
+L<XML::SAX::Expat|XML::SAX::Expat> and always specify the parser you wish to
+use by setting $XML::SAX::ParserPackage or the MAC_PROPERTYLIST_SAX_PARSER
+environment variable (see L</"DESCRIPTION">).
+
+Before version 0.85, this module contained a bug that caused double encoding of
+special characters as X[HT]ML entities. Thanks to Bion Pohl and
+L<http://ingz-inc.com/> for reporting this issue and supplying a fixed version.
+The implementation of the C<$ENCODE_ENTITIES> variable and the addition of the
+C<$ENCODE_UNSAFE_CHARS> variable are also due to Bion Pohl and / or
+L<http://ingz-inc.com/>.
 
 =head1 SUPPORT
 
@@ -413,16 +470,18 @@ Darren M. Kulp, C<< <kulp @ cpan.org> >>
 
 =head1 THANKS
 
-brian d foy, who created the L<Mac::PropertyList> module whose tests were
-appropriated for this module.
+brian d foy, who created the L<Mac::PropertyList|Mac::PropertyList> module
+whose tests were appropriated for this module.
+
+Bion Pohl and L<http://ingz-inc.com>, for bug report and patch submission.
 
 =head1 SEE ALSO
 
-L<Mac::PropertyList>, the inspiration for this module.
+L<Mac::PropertyList|Mac::PropertyList>, the inspiration for this module.
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2007 by Darren Kulp
+Copyright (C) 2007-2010 by Darren Kulp
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.4 or,
